<?php declare(strict_types=1);

namespace Drupal\billwerk_subscriptions\Form;

use Drupal\billwerk_subscriptions\Subscriber;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a Billwerk Subscriptions form.
 */
final class BillwerkUserRefreshSubscriptionForm extends FormBase {

	public function __construct(
		protected readonly Api $api
	) {
	}

	/**
	 * {@inheritdoc}
	 */
	public static function create(ContainerInterface $container) {
		// Instantiates this form class.
		return new static(
			// Load the service required to construct this class.
			$container->get('billwerk_subscriptions.api')
		);
	}

	/**
	 * {@inheritdoc}
	 */
	public function getFormId(): string {
		return 'billwerk_subscriptions_billwerk_user_refresh_subscription';
	}

	/**
	 * {@inheritdoc}
	 */
	public function buildForm(array $form, FormStateInterface $form_state, AccountInterface $user = NULL): array {
		if ($user) {
			$form_state->set('user', $user);

			$subscriber = Subscriber::load($user);
			if (!$subscriber->hasUserBillwerkContractId()) {
				$form['message'] = [
					'#type' => 'inline_template',
					'#template' => '<div>This user does not have any Billwerk Contract ID set. Assign a Billwerk Contract ID and vice-versa add the Drupal user id (UID) in the Billwerk: "External Customer ID" field to connect the user account from both sides.</div>',
					'#context' => [],
				];
				return $form;
			}

			$form['intro'] = [
				'#type' => 'container',
				'intro' => [
					'#prefix' => '<p><strong>',
					'#suffix' => '</strong></p>',
					'#markup' => $this->t('This action will <em>not</em> change any data at Billwerk. Only the Drupal user profile will be updated to match the current subscription.'),
				],
				'description' => [
					'#prefix' => '<p>',
					'#suffix' => '</p>',
					'#markup' => $this->t('By pressing the "Refresh user subscription"-button, %user\'s (UID %uid) data will be refreshed from Billwerk. Applying the latest state from Billwerk Contract that can be associated with this user (e.g. active subscription status, user data and respective roles). What exactly happens is subject to the subscription handler implementation used.', [
						'%uid' => $user->id(),
						'%user' => $user->getDisplayName(),
					]),
				],
			];

			$form['contract_information'] = [
				'#type' => 'inline_template',
				'#template' => '<div>This is the current Billwerk Customer and Contract information for this user:</div>
          <dl>
            <dt>Environment:</dt>
            <dd>@envionment</dd>
            <dt>Customer:</dt>
            <dd><pre><code>@customerData</code></pre></dd>
            <dt>Contract:</dt>
            <dd><pre><code>@contractData</code></pre></dd>
          </dl>',
				'#context' => [
					'@environment' => $this->api->getEnvironment()->getEnvironment(),
					'@customerData' => $subscriber->getBillwerkCustomer()->toJson(),
					'@contractData' => $subscriber->getBillwerkContract()->toJson(),
				],
			];

			$form['actions'] = [
				'#type' => 'actions',
				'submit' => [
					'#type' => 'submit',
					'#value' => $this->t('Refresh user subscription'),
				],
			];
		}

		return $form;
	}

	/**
	 * {@inheritdoc}
	 */
	public function submitForm(array &$form, FormStateInterface $form_state): void {
		$user = $form_state->get('user');
		$subscriber = Subscriber::load($user);
		$subscriber->refreshFromBillwerkContractSubscription();
		$this->messenger()->addStatus($this->t(
			'%user\'s subscription has been refreshed.',
			[
				'%user' => $user->getDisplayName()
			]
    ));
	}

	/**
	 * Checks access for a specific request.
	 *
	 * @param \Drupal\Core\Session\AccountInterface $account
	 *   Run access checks for this account.
	 *
	 * @return \Drupal\Core\Access\AccessResultInterface
	 *   The access result.
	 */
	public function access(AccountInterface $account) {
		return AccessResult::allowedIfHasPermission($account, 'billwerk_subscriptions_manual_refresh_user_subscriptions')
			->andIf(AccessResult::allowedIf(Subscriber::loadByDrupalUid($account->id())->hasUserBillwerkContractId()));
	}

}
