<?php

declare(strict_types=1);

namespace Drupal\billwerk_subscriptions;

use Drupal\billwerk_subscriptions\DataObject\BillwerkContract;
use Drupal\billwerk_subscriptions\DataObject\BillwerkCustomer;
use Drupal\billwerk_subscriptions\Event\SubscriberOrderCreateEvent;
use Drupal\billwerk_subscriptions\Event\SubscriberRefreshUserEvent;
use Drupal\billwerk_subscriptions\Exception\SubscriberException;
use Drupal\Component\Serialization\Json;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Symfony\Component\EventDispatcher\EventDispatcher;

/**
 * Wraps a Drupal user entity and adds subscription-related functionality to this "Subscriber".
 *
 * This makes it easier to interact with subscription functionalities for this
 * user.
 *
 */
class Subscriber {

	const USER_FIELD_BILLWERK_CONTRACT_ID = 'field_billwerk_contract_id';

	protected readonly Api $api;

	protected readonly EventDispatcher $eventDispatcher;

	protected readonly BillwerkDataObjectFactory $billwerkDataObjectFactory;

	protected readonly LogHelper $logHelper;

	/**
	 * Constructs a Subscriber object.
	 */
	protected function __construct(
		protected readonly UserInterface $user,
		protected ?BillwerkContract $billwerkContract,
	) {
		// Initialize services (we should, but can't use DI here):
		$this->api = \Drupal::service("billwerk_subscriptions.api");
		$this->eventDispatcher = \Drupal::service("event_dispatcher");
		$this->billwerkDataObjectFactory = \Drupal::service("billwerk_subscriptions.billwerk_data_object_factory");
		$this->logHelper = \Drupal::service("billwerk_subscriptions.log_helper");
	}

	/**
	 * Load the subscriber by the Drupal User object.
	 *
	 * @param \Drupal\user\UserInterface $user
	 *
	 * @return \Drupal\billwerk_subscriptions\Subscriber
	 */
	public static function load(UserInterface $user): self {
		// Lazy-initialize the Subscriber, only initialize the contract if needed.
		// This allows us to use Subscriber for users before they have a subscription
		// and only fetch the subscription data if needed.
		return new self($user, NULL);
	}

	/**
	 * Load the subscriber by its Drupal UID (ExternalCustomerId at Billwerk).
	 *
	 * @param int $uid
	 *
	 * @return \Drupal\billwerk_subscriptions\Subscriber
	 */
	public static function loadByDrupalUid(int $uid): ?self {
		$user = User::load($uid);
		if ($user !== NULL) {
			return self::load($user);
		}
		else {
			return NULL;
		}
	}


	/**
	 * Load the subscriber by the Drupal user's email address.
	 *
	 * @param string $mail
	 * @return Subscriber|null
	 */
	public static function loadByMail(string $mail): ?self {
		$user = user_load_by_mail($mail);
		if ($user !== NULL) {
			return self::load($user);
		}
		else {
			return NULL;
		}
	}


	/**
	 * Load the subscriber by the Billwerk Contract ID.
	 *
	 * This looks up the Drupal user with the given $billwerkContractId in
	 * his user profile.
	 *
	 * @param mixed $billwerkContractId
	 * @return Subscriber|null
	 */
	public static function loadByContractId($billwerkContractId): ?self {
		$users = \Drupal::entityTypeManager()
			->getStorage('user')
			->loadByProperties([self::USER_FIELD_BILLWERK_CONTRACT_ID => $billwerkContractId]);
		$user = reset($users);

		if (!empty($user)) {
			/** @var UserInterface $user */
			return self::load($user);
		}
		else {
			return NULL;
		}
	}


	/**
	 * Refreshes the user account details from his Billwerk Contract Subscription.
	 *
	 * If the user has no Billwerk Contract ID assigned, no refresh will be
	 * performed and the Event won't be called.
	 * Use Subscriber::hasUserBillwerkContractId() before to check if the user
	 * has a Billwerk Contract ID set, if needed.
	 *
	 * Doesn't change anything on his own (to not make assumptions), but instead
	 * dispatches the  SubscriberRefreshFromBillwerkContractSubscriptionEvent
	 * so that handlers can implement what happens if the subscription needs
	 * to be updated.
	 *
	 * @return void
	 */
	public function refreshFromBillwerkContractSubscription(): void {
		if (!$this->hasUserBillwerkContractId()) {
			// Only process users with a Billwerk Contract ID set. Skip others!
			return;
		}

		// Simply trigger the event to let handlers implement what should happen
		// and pass the subscriber object:
		$subscriberRefreshSubscriptionEvent = new SubscriberRefreshUserEvent($this);
		$this->eventDispatcher->dispatch($subscriberRefreshSubscriptionEvent);
	}

	/**
	 * Creates a new Billwerk Contract and Customer by an order.
	 *
	 * @throws \Drupal\billwerk_subscriptions\Exception\SubscriberException
	 * @return \Drupal\billwerk_subscriptions\Subscriber
	 */
	public function billwerkCreateContract(): Subscriber {
		// Prefill the $orderData with typical defaults:
		// They can be altered in the Event individually.
		$email = $this->getUser()->getEmail();
		$uid = $this->getUser()->id();
		$language = $this->getUser()->getPreferredLangcode();
		$orderData = [
			'Cart' => [
				// This needs to be overwritten by the subscriber!
				'planVariantId' => NULL,
			],
			'Customer' => [
				// Other example values:
				// FirstName: response.firstName,
				// LastName: response.lastName,
				// Language parameter seems deprecated in favor of Locale?:
				// 'Language' => $language,
				'EmailAddress' => $email,
				// AdditionalEmailAddresses: [],
				// Address: {
				//   Street: 'TRIAL',
				//   HouseNumber: 'TRIAL',
				//   PostalCode: 'TRIAL',
				//   City: 'TRIAL',
				//   CompanyName: 'TRIAL',
				//   Country: 'DE',
				// },
				// AdditionalAddresses: [],
				'Locale' => $language,
				'DefaultBearerMedium' => 'Email',
				'ExternalCustomerId' => $uid,
				// Leave a note in Billwerk to show the origin of this user:
				'Note' => 'Created by Drupal billwerk_subscriptions module on Drupal User #' . $uid . ' creation at ' . date("c"),
			],
		];

		// The SubscriberOrderCreateEvent can be used to modify the order data individually.
		$subscriberOrderCreateEvent = new SubscriberOrderCreateEvent($orderData, $this);
		$this->eventDispatcher->dispatch($subscriberOrderCreateEvent);

		// Retrieve the possibly modified order data:
		$orderData = $subscriberOrderCreateEvent->getOrderData();
		$order = $this->api->createOrder($orderData);

		if ($subscriberOrderCreateEvent->getCommitOrder()) {
			$committedOrder = $this->api->commitOrder($order->Id, []);
		}
		$contractId = $committedOrder->Id;
		if (!empty($contractId)) {
			$this->setUserBillwerkContractId($contractId);
		}
		else {
			$this->logHelper->error('Registration for %email was not possible, Billwerk returned the following response, without the required order ID: %response', [
				'%email' => $email,
				'%response' => Json::encode($order),
			]);
			throw new SubscriberException('Contract ID could not be determined for committed order. Contract was created, but not assigned to the user "' . $this->getUser()->id() . '"');
		}
		return $this;
	}

	/**
	 * Deletes the customer account and its contracts at Billwerk entirely.
	 *
	 * @throws \Drupal\billwerk_subscriptions\Exception\SubscriberException
	 * @return void
	 */
	public function billwerkDeleteCustomer(): void {
		$billwerkCustomerId = $this->getBillwerkCustomer()->getId();
		if (!empty($billwerkCustomerId)) {
			$this->api->deleteCustomer($billwerkCustomerId);
			$this->setUserBillwerkContractId('');
		}
		else {
			throw new SubscriberException("Customer ID could not be determined for User #{$this->getUser()->id()}");
		}
	}

  /**
   * Sets the Billwerk Customer ExternalCustomerId to the Subscriber's user ID.
   *
   * @throws \Drupal\billwerk_subscriptions\Exception\SubscriberException
   * @return void
   */
	public function billwerkSetCustomerExternalCustomerId(): void {
		if (empty($this->getBillwerkCustomer())) {
			throw new SubscriberException("Customer could not be loaded for User #{$this->getUser()->id()}");
		}
		if ($this->getBillwerkCustomer()->getExternalCustomerId() == $this->getUser()->id()) {
			// The external customer ID already equals the Drupal user id.
			return;
		}
		// Set the Billwerk ExternalCustomerId to the user ID.
		$this->api->patchCustomer($this->getBillwerkCustomer()->getId(), ['ExternalCustomerId' => $this->getUser()->id()]);
	}

	/**
	 * Returns the BillwerkContract with lazy-loading.
	 *
	 *
	 *
	 * @return \Drupal\billwerk_subscriptions\DataObject\BillwerkContract
	 */
	public function getBillwerkContract(): ?BillwerkContract {
		if ($this->billwerkContract !== NULL) {
			return $this->billwerkContract;
		}
		else {
			if ($this->hasUserBillwerkContractId()) {
				$billwerkContractId = $this->getUserBillwerkContractId();
				$this->billwerkContract = $this->billwerkDataObjectFactory->billwerkLoadBillwerkContract($billwerkContractId);
			}
		}

		return $this->billwerkContract;
	}

	/**
	 * Summary of getCustomerData.
	 *
	 * @return ?\Drupal\billwerk_subscriptions\DataObject\BillwerkCustomer
	 */
	public function getBillwerkCustomer(): ?BillwerkCustomer {
		return $this->getBillwerkContract()->getBillwerkCustomer();
	}

	/**
	 * Returns the user.
	 *
	 * @return \Drupal\user\UserInterface
	 */
	public function getUser(): UserInterface {
		// Safety net to never accidentially return the anonymous user:
		if ($this->user->isAnonymous()) {
			throw new SubscriberException('Subscriber should never be anonymous! Something is wrong.');
		}
		return $this->user;
	}


	/**
	 * Sets the Billwerk Contract ID in the Drupal user profile.
	 *
	 * @param string $billwerkContractId
	 * @param bool $save
	 * @return void
	 */
	public function setUserBillwerkContractId(string $billwerkContractId, bool $save = TRUE): void {
		$this->user->set(self::USER_FIELD_BILLWERK_CONTRACT_ID, $billwerkContractId);
		if ($save) {
			$this->user->save();
		}
	}

	/**
	 * Returns the Billwerk Contract ID stored in the Drupal user profile. If none ist stored, returns NULL.
	 *
	 * @return ?string
	 */
	public function getUserBillwerkContractId(): ?string {
		return $this->user->get(self::USER_FIELD_BILLWERK_CONTRACT_ID)->getString() ?: NULL;
	}

	/**
	 * Returns true if the user has a Billwerk Contract ID stored, else false.
	 *
	 * @return bool
	 */
	public function hasUserBillwerkContractId(): bool {
		$userBillwerkContractId = $this->getUserBillwerkContractId();
		return !empty($userBillwerkContractId);
	}

}
